#include <stdio.h>
#include <stdlib.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <dlfcn.h>		// dlopen()

static int use_xkblib = -1;
static Display* fl_display;

#ifdef HAVE_XKB

#include <X11/XKBlib.h>

static KeySym fl_KeycodeToKeysymA(Display *d, KeyCode k, unsigned i, unsigned l) {
  KeySym ks;
  if (use_xkblib < 0) { // this is done only once
    use_xkblib = 0;
    int libmajor = XkbMajorVersion, libminor = XkbMinorVersion;
    printf ("[ 2] libmajor, libminor = %d,%d\n",libmajor,libminor);
    int kb_op_code, kb_event_code, kb_err_code;
    bool doit = XkbLibraryVersion(&libmajor, &libminor);
    printf ("[ 3] XkbLibraryVersion(&libmajor, &libminor) = %d\n",doit);
    printf ("[ 4] libmajor, libminor = %d,%d\n",libmajor,libminor);
    if (doit) doit = XkbQueryExtension(fl_display, &kb_op_code, &kb_event_code, &kb_err_code, &libmajor, &libminor);
    printf ("[ 5] XkbQueryExtension(...) = %d\n",doit);
    printf ("[ 6] libmajor, libminor = %d,%d\n",libmajor,libminor);
    if (doit) use_xkblib = 1;
  }
  if (use_xkblib > 0) {
    ks = XkbKeycodeToKeysym(fl_display, k, i, l);
    printf ("[ 7] KeySym = %d\n",(int)ks);
    // return ks;
  }
  ks = XKeycodeToKeysym(d, k, i);
  printf ("[ 8] KeySym = %d\n",(int)ks);
  return ks;
}

#endif

static KeySym fl_KeycodeToKeysymB(Display *d, KeyCode k, unsigned i, unsigned l) {
  typedef KeySym (*XkbKeycodeToKeysym_type)(Display *d, KeyCode k, unsigned i, unsigned l);
  typedef KeySym (*XKeycodeToKeysym_type)(Display *d, KeyCode k, unsigned i);
  static XkbKeycodeToKeysym_type XkbKeycodeToKeysym;
  static XKeycodeToKeysym_type XKeycodeToKeysym;
  if (use_xkblib < 0) { // this is done only once
    use_xkblib = 0;
    int kb_ext_code, kb_event_code, kb_err_code;
    int libmajor = 1, libminor = 0;
    typedef Bool (*XkbQueryExtension_type)(Display *, int*,int*,int*,int*,int*);

    void *handle = dlopen(NULL, RTLD_LAZY); // search symbols in executable
    XkbQueryExtension_type XkbQueryExtension_f = (XkbQueryExtension_type)dlsym(handle, "XkbQueryExtension");
    XkbKeycodeToKeysym = (XkbKeycodeToKeysym_type)dlsym(handle, "XkbKeycodeToKeysym");
    XKeycodeToKeysym = (XKeycodeToKeysym_type)dlsym(handle, "XKeycodeToKeysym");

    // make sure that the X server has the XKB extension
    if ( !( XkbQueryExtension_f && XkbKeycodeToKeysym &&
           XkbQueryExtension_f(fl_display, &kb_ext_code, &kb_event_code, &kb_err_code, &libmajor, &libminor) ) ) XkbKeycodeToKeysym = NULL;
    if (XkbKeycodeToKeysym) use_xkblib = 1;
    else XKeycodeToKeysym = (XKeycodeToKeysym_type)dlsym(handle, "XKeycodeToKeysym");
  }
  printf("[46] XkbKeycodeToKeysym = %p\n", XkbKeycodeToKeysym);
  printf("[47] XKeycodeToKeysym   = %p\n", XKeycodeToKeysym);
  if (use_xkblib > 0) return XkbKeycodeToKeysym(d, k, i, l);
  return XKeycodeToKeysym(d, k, i);
}

int main (int argc, char **argv) {
  KeyCode k = 11;
  if (argc > 1) k = atoi(argv[1]);
  KeySym ks;
  unsigned i = 0, l = 0;
  fl_display = XOpenDisplay(":0.0");
  printf ("[ 1] fl_display = %p\n",fl_display);

#ifdef HAVE_XKB
  printf ("\n---- Calling fl_KeycodeToKeysymA ----\n");
  ks = fl_KeycodeToKeysymA(fl_display, k, i, l);
  printf ("[11] KeySym ks = %d\n", (int)ks);
#endif

  use_xkblib = -1; // reset flag for simulation
  printf ("\n---- Calling fl_KeycodeToKeysymB ----\n");
  ks = fl_KeycodeToKeysymB(fl_display, k, i, l);
  printf ("[12] KeySym ks = %d\n", (int)ks);
  return 0;
}

